import pytest
from starlette.testclient import TestClient

from fast_gov_uk import core


@pytest.fixture
def dev_client():
    settings = {
        "SERVICE_NAME": "Fast",
        "DATABASE_URL": ":memory:",
        "DEV_MODE": True,
        "NOTIFY_API_KEY": None
    }
    app = core.Fast(settings)
    return TestClient(app)


def test_demo(dev_client):
    response = dev_client.get("/demo")
    assert response.status_code == 200



import pytest

import fast_gov_uk.design_system as ds
from fast_gov_uk.design_system import Next, Previous


@pytest.mark.parametrize(
    "kwargs, expected",
    (
        (
            {"text": "Test", "href": "/test"},
            '<a href="/test" class="govuk-back-link">Test</a>',
        ),
        (
            {"text": "Test", "href": "/test", "inverse": True},
            '<a href="/test" class="govuk-back-link govuk-back-link--inverse">Test</a>',
        ),
        (
            {"text": "", "href": "/test"},
            '<a href="/test" class="govuk-back-link"></a>',
        ),
        (
            {
                "href": "/test",
            },
            '<a href="/test" class="govuk-back-link">Back</a>',
        ),
    ),
)
def test_backlink(kwargs, expected, html):
    """Test backlink with various parameters.
    Args:
        kwargs (dict): The arguments to pass to backlink.
        expected (str): The expected HTML output.
    """
    backlink = ds.BackLink(**kwargs)
    assert html(backlink) == html(expected)


@pytest.mark.parametrize(
    "kwargs, expected",
    (
        (
            {"text": "Test", "href": "/test"},
            '<a href="/test" data-module="govuk-skip-link" class="govuk-skip-link">Test</a>',
        ),
    ),
)
def test_skip_link(kwargs, expected, html):
    """Test SkipLink with various parameters.
    Args:
        kwargs (dict): The arguments to pass to SkipLink.
        expected (str): The expected HTML output.
    """
    link = ds.SkipLink(**kwargs)
    assert html(link) == html(expected)


@pytest.mark.parametrize(
    "args, expected",
    (
        (
            [("Home", "/"), ("Test", "/test")],
            (
                '<nav aria-label="Breadcrumb" class="govuk-breadcrumbs">'
                    '<ol class="govuk-breadcrumbs__list">'
                        '<li class="govuk-breadcrumbs__list-item">'
                            '<a href="/" class="govuk-breadcrumbs__link">Home</a>'
                        "</li>"
                        '<li class="govuk-breadcrumbs__list-item">'
                            '<a href="/test" class="govuk-breadcrumbs__link">Test</a>'
                        "</li>"
                    "</ol>"
                "</nav>"
            ),
        ),
    ),
)
def test_breadcrumb(args, expected, html):
    """Test Breadcrumb with various parameters.
    Args:
        args (list): The args to pass to Breadcrumb.
        expected (str): The expected HTML output.
    """
    bc = ds.Breadcrumbs(*args)
    assert html(bc) == html(expected)


@pytest.mark.parametrize(
    "kwargs, expected",
    (
        (
            {},
            (
                '<div data-module="govuk-exit-this-page" class="govuk-exit-this-page">'
                    '<a href="https://www.bbc.co.uk/weather" role="button" draggable="false" data-module="govuk-button" rel="nofollow noreferrer" class="govuk-button govuk-button--warning govuk-exit-this-page__button govuk-js-exit-this-page-button">'
                        '<span class="govuk-visually-hidden">Emergency</span>'
                        "Exit this page"
                    "</a>"
                "</div>"
            ),
        ),
        (
            {"text": "Test", "href": "/test"},
            (
                '<div data-module="govuk-exit-this-page" class="govuk-exit-this-page">'
                    '<a href="/test" role="button" draggable="false" data-module="govuk-button" rel="nofollow noreferrer" class="govuk-button govuk-button--warning govuk-exit-this-page__button govuk-js-exit-this-page-button">'
                        '<span class="govuk-visually-hidden">Emergency</span>'
                        "Test"
                    "</a>"
                "</div>"
            ),
        ),
    ),
)
def test_exit_page(kwargs, expected, html):
    """Test Exit Page with various parameters.
    Args:
        kwargs (dict): The arguments to pass to ExitPage.
        expected (str): The expected HTML output.
    """
    exit = ds.ExitPage(**kwargs)
    assert html(exit) == html(expected)


@pytest.mark.parametrize(
    "kwargs, expected",
    (
        (
            {"text": "test", "href": "/test"},
            (
                '<li class="govuk-service-navigation__item">'
                    '<a href="/test" class="govuk-service-navigation__link">test</a>'
                "</li>"
            ),
        ),
        (
            {"text": "test", "href": "/test", "active": True},
            (
                '<li class="govuk-service-navigation__item govuk-service-navigation__item--active">'
                    '<a href="/test" aria-current="true" class="govuk-service-navigation__link">'
                        '<strong class="govuk-service-navigation__active-fallback">test</strong>'
                    "</a>"
                "</li>"
            ),
        ),
    ),
)
def test_navigation_link(kwargs, expected, html):
    """Test NavigationLink with various parameters.
    Args:
        kwargs (dict): The arguments to pass to NvigationLink.
        expected (str): The expected HTML output.
    """
    link = ds.NavigationLink(**kwargs)
    assert html(link) == html(expected)


@pytest.mark.parametrize(
    "args, expected",
    (
        (
            [
                ds.NavigationLink("Test Label 1", "/test1"),
                ds.NavigationLink("Test Label 2", "/test2"),
            ],
            (
                '<section aria-label="Service information" data-module="govuk-service-navigation" class="govuk-service-navigation">'
                    '<div class="govuk-width-container">'
                        '<div class="govuk-service-navigation__container">'
                            '<span class="govuk-service-navigation__service-name">'
                                '<a href="/" class="govuk-service-navigation__link">Test</a>'
                            "</span>"
                            '<nav aria-label="Menu" class="govuk-service-navigation__wrapper">'
                                '<button type="button" aria-controls="navigation" hidden class="govuk-service-navigation__toggle govuk-js-service-navigation-toggle">Menu</button>'
                                '<ul id="navigation" class="govuk-service-navigation__list">'
                                    '<li class="govuk-service-navigation__item">'
                                        '<a href="/test1" class="govuk-service-navigation__link">Test Label 1</a>'
                                    "</li>"
                                    '<li class="govuk-service-navigation__item">'
                                        '<a href="/test2" class="govuk-service-navigation__link">Test Label 2</a>'
                                    "</li>"
                                "</ul>"
                            "</nav>"
                        "</div>"
                    "</div>"
                "</section>"
            ),
        ),
    ),
)
def test_navigation(args, expected, html):
    """Test Navigation with various parameters.
    Args:
        args (list): The arguments to pass to Navigation.
        expected (str): The expected HTML output.
    """
    nav = ds.Navigation(*args, service_name="Test")
    assert html(nav) == html(expected)


@pytest.mark.parametrize(
    "kwargs, expected",
    (
        (
            {
                "links":
                    [
                        ds.PaginationLink("1", "/test?page=1", active=True),
                        ds.PaginationLink("2", "/test?page=2"),
                        ds.PaginationLink("3", "/test?page=3"),
                    ],
                    "next_link": "/test?page=2",
            },
            (
                '<nav aria-label="Pagination" class="govuk-pagination">'
                    '<ul class="govuk-pagination__list">'
                        '<li class="govuk-pagination__item govuk-pagination__item--current">'
                            '<a href="/test?page=1" aria-label="Page 1" aria-current="page" class="govuk-link govuk-pagination__link">1</a>'
                        "</li>"
                        '<li class="govuk-pagination__item">'
                            '<a href="/test?page=2" aria-label="Page 2" class="govuk-link govuk-pagination__link">2</a>'
                        "</li>"
                        '<li class="govuk-pagination__item">'
                            '<a href="/test?page=3" aria-label="Page 3" class="govuk-link govuk-pagination__link">3</a>'
                        "</li>"
                    "</ul>"
                    '<div class="govuk-pagination__next">'
                        '<a href="/test?page=2" rel="next" class="govuk-link govuk-pagination__link">'
                            '<span class="govuk-pagination__link-title">Next<span class="govuk-visually-hidden"> page</span></span>'
                            f"{Next()}"
                        "</a>"
                    "</div>"
                "</nav>"
            ),
        ),
        (
            {
                "links":
                    [
                        ds.PaginationLink("1", "/test?page=1"),
                        ds.PaginationLink("2", "/test?page=2", active=True),
                        ds.PaginationLink("3", "/test?page=3"),
                    ],
                    "prev_link": "/test?page=1",
                    "next_link": "/test?page=3",
            },
            (
                '<nav aria-label="Pagination" class="govuk-pagination">'
                    '<div class="govuk-pagination__prev">'
                        '<a href="/test?page=1" rel="prev" class="govuk-link govuk-pagination__link">'
                            f"{Previous()}"
                            '<span class="govuk-pagination__link-title">Previous<span class="govuk-visually-hidden"> page</span></span>'
                        "</a>"
                    "</div>"
                    '<ul class="govuk-pagination__list">'
                        '<li class="govuk-pagination__item">'
                            '<a href="/test?page=1" aria-label="Page 1" class="govuk-link govuk-pagination__link">1</a>'
                        "</li>"
                        '<li class="govuk-pagination__item govuk-pagination__item--current">'
                            '<a href="/test?page=2" aria-label="Page 2" aria-current="page" class="govuk-link govuk-pagination__link">2</a>'
                        "</li>"
                        '<li class="govuk-pagination__item">'
                            '<a href="/test?page=3" aria-label="Page 3" class="govuk-link govuk-pagination__link">3</a>'
                        "</li>"
                    "</ul>"
                    '<div class="govuk-pagination__next">'
                        '<a href="/test?page=3" rel="next" class="govuk-link govuk-pagination__link">'
                        '<span class="govuk-pagination__link-title">Next<span class="govuk-visually-hidden"> page</span></span>'
                            f"{Next()}"
                        "</a>"
                    "</div>"
                "</nav>"
            ),
        ),
        (
            {
                "links":
                    [
                        ds.PaginationLink("1", "/test?page=1"),
                        ds.PaginationLink("2", "/test?page=2"),
                        ds.PaginationLink("3", "/test?page=3", active=True),
                    ],
                    "prev_link": "/test?page=2",
            },
            (
                '<nav aria-label="Pagination" class="govuk-pagination">'
                    '<div class="govuk-pagination__prev">'
                        '<a href="/test?page=2" rel="prev" class="govuk-link govuk-pagination__link">'
                            f"{Previous()}"
                            '<span class="govuk-pagination__link-title">Previous<span class="govuk-visually-hidden"> page</span></span>'
                        "</a>"
                    "</div>"
                    '<ul class="govuk-pagination__list">'
                        '<li class="govuk-pagination__item">'
                            '<a href="/test?page=1" aria-label="Page 1" class="govuk-link govuk-pagination__link">1</a>'
                        "</li>"
                        '<li class="govuk-pagination__item">'
                            '<a href="/test?page=2" aria-label="Page 2" class="govuk-link govuk-pagination__link">2</a>'
                        "</li>"
                        '<li class="govuk-pagination__item govuk-pagination__item--current">'
                            '<a href="/test?page=3" aria-label="Page 3" aria-current="page" class="govuk-link govuk-pagination__link">3</a>'
                        "</li>"
                    "</ul>"
                "</nav>"
            ),
        ),
    ),
)
def test_pagination(kwargs, expected, html):
    """Test Pagination with various parameters.
    Args:
        kwargs (dict): The kwargs to pass to Pagination.
        expected (str): The expected HTML output.
    """
    links = kwargs.pop("links")
    page = ds.Pagination(*links, **kwargs)
    assert html(page) == html(expected)


@pytest.mark.parametrize("component", (
    ds.BackLink("/", hx_test="foo"),
    ds.SkipLink("/", hx_test="foo"),
    ds.Breadcrumbs(("test", "/"), hx_test="foo"),
    ds.ExitPage(hx_test="foo"),
    ds.NavigationLink("test", "/", hx_test="foo"),
    ds.Navigation(hx_test="foo"),
    ds.Pagination(hx_test="foo"),
))
def test_html_attribute(component, html):
    """
    Test that passes an html attribute to components - that is not
    explicitly handled but should be passed through to the
    underlying FT.
    """
    assert 'hx-test="foo"' in html(component)



import fast_gov_uk.design_system as ds


def test_backlink(html):
    """
    Link: https://design-system.service.gov.uk/components/back-link/
    """
    backlink = ds.BackLink("#")
    assert html(backlink) == html(
        '<a href="#" class="govuk-back-link">' \
            'Back'
        '</a>'
    )


def test_backlink_dark(html):
    """
    Link: https://design-system.service.gov.uk/components/back-link/
    """
    backlink = ds.BackLink("#", inverse=True)
    assert html(backlink) == html(
        '<a href="#" class="govuk-back-link govuk-back-link--inverse">'
            'Back'
        '</a>'
    )


def test_skiplink(html):
    """
    Link: https://design-system.service.gov.uk/components/skip-link/
    """
    skiplink = ds.SkipLink("#")
    assert html(skiplink) == html(
        '<a href="#" class="govuk-skip-link" data-module="govuk-skip-link">'
            'Skip to main content'
        '</a>'
    )


def test_breadcrumbs(html):
    """
    Link: https://design-system.service.gov.uk/components/breadcrumbs/
    """
    breadcrumbs = ds.Breadcrumbs(
        ("Home", "#"),
        ("Passports, travel and living abroad", "#"),
        ("Travel abroad", "#"),
    )
    assert html(breadcrumbs) == html(
        '<nav class="govuk-breadcrumbs" aria-label="Breadcrumb">'
        '<ol class="govuk-breadcrumbs__list">'
            '<li class="govuk-breadcrumbs__list-item">'
                '<a class="govuk-breadcrumbs__link" href="#">Home</a>'
            '</li>'
            '<li class="govuk-breadcrumbs__list-item">'
                '<a class="govuk-breadcrumbs__link" href="#">Passports, travel and living abroad</a>'
            '</li>'
            '<li class="govuk-breadcrumbs__list-item">'
                '<a class="govuk-breadcrumbs__link" href="#">Travel abroad</a>'
            '</li>'
        '</ol>'
        '</nav>'
    )


def test_breadcrumbs_collpse(html):
    """
    Link: https://design-system.service.gov.uk/components/breadcrumbs/
    """
    breadcrumbs = ds.Breadcrumbs(
        ("Home", "#"),
        ("Environment", "#"),
        ("Rural and countryside", "#"),
        ("Rural development and land management", "#"),
        ("Economic growth in rural areas", "#"),
        collapse_on_mobile=True,
    )
    assert html(breadcrumbs) == html(
        '<nav class="govuk-breadcrumbs govuk-breadcrumbs--collapse-on-mobile" aria-label="Breadcrumb">'
        '<ol class="govuk-breadcrumbs__list">'
            '<li class="govuk-breadcrumbs__list-item">'
                '<a class="govuk-breadcrumbs__link" href="#">Home</a>'
            '</li>'
            '<li class="govuk-breadcrumbs__list-item">'
                '<a class="govuk-breadcrumbs__link" href="#">Environment</a>'
            '</li>'
            '<li class="govuk-breadcrumbs__list-item">'
                '<a class="govuk-breadcrumbs__link" href="#">Rural and countryside</a>'
            '</li>'
            '<li class="govuk-breadcrumbs__list-item">'
                '<a class="govuk-breadcrumbs__link" href="#">Rural development and land management</a>'
            '</li>'
            '<li class="govuk-breadcrumbs__list-item">'
                '<a class="govuk-breadcrumbs__link" href="#">Economic growth in rural areas</a>'
            '</li>'
        '</ol>'
        '</nav>'
    )


def test_exit_page(html):
    """
    Link: https://design-system.service.gov.uk/components/exit-this-page/
    """
    exit_page = ds.ExitPage()
    assert html(exit_page) == html(
        '<div class="govuk-exit-this-page" data-module="govuk-exit-this-page">'
        '<a href="https://www.bbc.co.uk/weather" role="button" draggable="false" class="govuk-button govuk-button--warning govuk-exit-this-page__button govuk-js-exit-this-page-button" data-module="govuk-button" rel="nofollow noreferrer">'
            '<span class="govuk-visually-hidden">Emergency</span> Exit this page'
        '</a>'
        '</div>'
    )


def test_navigation(html):
    """
    Link: https://design-system.service.gov.uk/components/service-navigation/
    """
    navigation = ds.Navigation(
        ds.NavigationLink("Navigation item 1", "#"),
        ds.NavigationLink("Navigation item 2", "#", active=True),
        ds.NavigationLink("Navigation item 3", "#"),
        service_name="Service name",
    )
    assert html(navigation) == html(
        '<section aria-label="Service information" class="govuk-service-navigation"'
        'data-module="govuk-service-navigation">'
        '<div class="govuk-width-container">'
            '<div class="govuk-service-navigation__container">'
            '<span class="govuk-service-navigation__service-name">'
                # '<a href="#" class="govuk-service-navigation__link">Service name</a>'
                # The spec had href for service link that was "#" instead of the root
                '<a href="/" class="govuk-service-navigation__link">Service name</a>'
            '</span>'
            '<nav aria-label="Menu" class="govuk-service-navigation__wrapper">'
                '<button type="button" class="govuk-service-navigation__toggle govuk-js-service-navigation-toggle" aria-controls="navigation" hidden>'
                'Menu'
                '</button>'
                '<ul class="govuk-service-navigation__list" id="navigation">'
                '<li class="govuk-service-navigation__item">'
                    '<a class="govuk-service-navigation__link" href="#">Navigation item 1</a>'
                '</li>'
                '<li class="govuk-service-navigation__item govuk-service-navigation__item--active">'
                    '<a class="govuk-service-navigation__link" href="#" aria-current="true">'
                        '<strong class="govuk-service-navigation__active-fallback">Navigation item 2</strong>'
                    '</a>'
                '</li>'
                '<li class="govuk-service-navigation__item">'
                    '<a class="govuk-service-navigation__link" href="#">Navigation item 3</a>'
                '</li>'
                '</ul>'
            '</nav>'
            '</div>'
        '</div>'
        '</section>'
    )


def test_pagination(html):
    """
    Link: https://design-system.service.gov.uk/components/pagination/
    """
    pagination = ds.Pagination(
        ds.PaginationLink("1", "#"),
        ds.PaginationLink("2", "#", active=True),
        ds.PaginationLink("3", "#"),
        prev_link="#",
        next_link="#",
    )
    assert html(pagination) == html(
        '<nav class="govuk-pagination" aria-label="Pagination">'
            '<div class="govuk-pagination__prev">'
                '<a class="govuk-link govuk-pagination__link" href="#" rel="prev">'
                    '<svg class="govuk-pagination__icon govuk-pagination__icon--prev" xmlns="http://www.w3.org/2000/svg" height="13" width="15" aria-hidden="true" focusable="false" viewBox="0 0 15 13">'
                        '<path d="m6.5938-0.0078125-6.7266 6.7266 6.7441 6.4062 1.377-1.449-4.1856-3.9768h12.896v-2h-12.984l4.2931-4.293-1.414-1.414z"></path>'
                    '</svg>'
                    '<span class="govuk-pagination__link-title">'
                        'Previous<span class="govuk-visually-hidden"> page</span>'
                    '</span>'
                '</a>'
            '</div>'
            '<ul class="govuk-pagination__list">'
                '<li class="govuk-pagination__item">'
                    '<a class="govuk-link govuk-pagination__link" href="#" aria-label="Page 1">1</a>'
                '</li>'
                '<li class="govuk-pagination__item govuk-pagination__item--current">'
                    '<a class="govuk-link govuk-pagination__link" href="#" aria-label="Page 2" aria-current="page">2</a>'
                '</li>'
                '<li class="govuk-pagination__item">'
                    '<a class="govuk-link govuk-pagination__link" href="#" aria-label="Page 3">3</a>'
                '</li>'
            '</ul>'
            '<div class="govuk-pagination__next">'
                '<a class="govuk-link govuk-pagination__link" href="#" rel="next">'
                    '<span class="govuk-pagination__link-title">'
                        'Next<span class="govuk-visually-hidden"> page</span>'
                    '</span>'
                    '<svg class="govuk-pagination__icon govuk-pagination__icon--next" xmlns="http://www.w3.org/2000/svg" height="13" width="15" aria-hidden="true" focusable="false" viewBox="0 0 15 13">'
                        '<path d="m8.107-0.0078125-1.4136 1.414 4.2926 4.293h-12.986v2h12.896l-4.1855 3.9766 1.377 1.4492 6.7441-6.4062-6.7246-6.7266z"></path>'
                    '</svg>'
                '</a>'
            '</div>'
        '</nav>'
    )


def test_pagination_block(html):
    """
    Link: https://design-system.service.gov.uk/components/pagination/
    """
    pagination = ds.PaginationBlock(
        ("Applying for a provisional lorry or bus licence", "#"),
        ("Driver CPC part 1 test: theory", "#"),
    )
    assert html(pagination) == html(
        '<nav class="govuk-pagination govuk-pagination--block" aria-label="Pagination">'
            '<div class="govuk-pagination__prev">'
                '<a class="govuk-link govuk-pagination__link" href="#" rel="prev">'
                    '<svg class="govuk-pagination__icon govuk-pagination__icon--prev" xmlns="http://www.w3.org/2000/svg" height="13" width="15" aria-hidden="true" focusable="false" viewBox="0 0 15 13">'
                        '<path d="m6.5938-0.0078125-6.7266 6.7266 6.7441 6.4062 1.377-1.449-4.1856-3.9768h12.896v-2h-12.984l4.2931-4.293-1.414-1.414z"></path>'
                    '</svg>'
                    '<span class="govuk-pagination__link-title">'
                        'Previous<span class="govuk-visually-hidden"> page</span>'
                    '</span>'
                    '<span class="govuk-visually-hidden">:</span>'
                    '<span class="govuk-pagination__link-label">Applying for a provisional lorry or bus licence</span>'
                '</a>'
            '</div>'
            '<div class="govuk-pagination__next">'
                '<a class="govuk-link govuk-pagination__link" href="#" rel="next">'
                    '<svg class="govuk-pagination__icon govuk-pagination__icon--next" xmlns="http://www.w3.org/2000/svg" height="13" width="15" aria-hidden="true" focusable="false" viewBox="0 0 15 13">'
                        '<path d="m8.107-0.0078125-1.4136 1.414 4.2926 4.293h-12.986v2h12.896l-4.1855 3.9766 1.377 1.4492 6.7441-6.4062-6.7246-6.7266z"></path>'
                    '</svg>'
                    '<span class="govuk-pagination__link-title">'
                        'Next<span class="govuk-visually-hidden"> page</span>'
                    '</span>'
                    '<span class="govuk-visually-hidden">:</span>'
                    '<span class="govuk-pagination__link-label">Driver CPC part 1 test: theory</span>'
                '</a>'
            '</div>'
        '</nav>'
    )



from os import environ as env

from fasthtml import common as fh

import fast_gov_uk.design_system as ds
from fast_gov_uk import forms
from fast_gov_uk.core import Fast

# This is your settings. They are set for development on your
# own computer by default. When you are deploying your service,
# you can override them by setting environment variables.
SETTINGS = {
    "SERVICE_NAME": env.get("SERVICE_NAME", "Fast Gov UK"),
    "DATABASE_URL": env.get("DATABASE_URL", "data/service.db"),
    "DEV_MODE": env.get("DEV_MODE", True),
    "NOTIFY_API_KEY": env.get("NOTIFY_API_KEY", None),
}


# This creates the Fast object that encapsulates everything
# in your service
fast = Fast(SETTINGS)


# If I do @fast.page() instead, this page will be available
# on /home i.e. the name of the function
@fast.page("/")
def home():
    return ds.Page(
        # A single Paragraph
        ds.P("Welcome to Fast Gov UK.")
    )


# I can do @fast.form("foo") instead to make this page available
# on /forms/foo instead of the default /forms/feedback
@fast.form
def feedback(data=None):
    """
    Feedback form is common and recommended in gov.uk services.
    This form also serves as an example for how to add new forms
    to fast-gov-uk.

    Forms can be empty when rendered for the first time. Forms
    can also have data e.g. when you enter invalid information in
    a field and the you get the form again with errors as well as
    what you filled in.

    In view of the above, this is a function that takes data as as
    argument and passes it into the Form.

    It also sets -

    success_url: Redirects here if the form is submitted correctly

    cta: Call to action - label for the form button

    db: This is a DBForm so it needs a reference to the database
    so that it can save upon correct submission

    P.S. This is an attempt to reproduce -
    https://www.gov.uk/service-manual/service-assessments/get-feedback-page
    """
    # A DBForm gets saved to the database when its valid
    return forms.Form(
        "feedback",
        ds.H1("Give feedback on Fast GOV UK"),
        ds.H2("Satisfaction survey"),
        ds.Radios(
            name="satisfaction",
            label="Overall, how satisfied did you feel about Fast Gov UK?",
            choices={
                "very-satisfied": "Very Satisfied",
                "satisfied": "Satisfied",
                "neutral": "Neither satisfied not dissatisfied",
                "dissatisfied": "Dissatisfied",
                "very-dissatisfied": "Very dissatisfied",
            },
            heading="s",
        ),
        ds.CharacterCount(
            name="comments",
            label="How could we improve this service?",
            maxchars=1200,
            required=False,
            hint=(
                "Do not include any personal or financial information, "
                "for example your national insurance number."
            ),
            heading="s",
        ),
        backends=[forms.DBBackend(db=fast.db)],
        success_url="/",
        data=data,
        cta="Send feedback",
        db=fast.db,
    )


@fast.wizard
def equality(step=0, data=None):
    """
    This is an example of a question-protocol aka wizard form.
    It steps through the fields one at a time.

    It takes step as an argument to know which field to show.
    It also takes data as an argument to fill in the fields
    that have already been filled in.

    Note that this form does not do anything with the data
    when it is completed. You can add your own processing logic
    in the process method of the Questions class.
    """
    return forms.Wizard(
        "equality",
        forms.Question(
            ds.H1("We have received your application"),
            ds.P(
                "Before you finish using the service, "
                "we'd like to ask some equality questions."
            ),
            ds.P(
                "[Add a couple of sentences explaining why "
                "you're asking the questions and what you'll "
                "do with the information]."
            ),
            ds.Radios(
                name="permission",
                label="Do you want to answer the equality questions?",
                choices={
                    "yes": "Yes, answer the equality questions",
                    "no": "No, skip the equality questions",
                },
                heading="m",
                hint="These questions are optional.",
            ),
            ds.Detail(
                "Why we ask equality questions",
                ds.P("[Consider adding an optional longer explanation]")
            ),
        ),
        forms.Question(
            ds.DateInput(
                name="dob",
                label="What is your date of birth?",
                hint=(
                    "For example, 31 3 1980 - if you prefer not to say, "
                    "continue without entering any information"
                ),
                heading="l",
            ),
            predicates={"permission": "yes"},
        ),
        forms.Question(
            ds.Radios(
                name="health",
                label=(
                    "Do you have any physical or mental health conditions or illness "
                    "lasting or expected to last 12 months or more?"
                ),
                choices={"yes": "Yes", "no": "No", "skip": "Prefer not to say"},
                heading="l",
            ),
            predicates={"permission": "yes"},
        ),
        forms.Question(
            ds.Radios(
                name="ability",
                label=(
                    "Do any of your conditions or illnesses reduce your ability "
                    "to carry out day to day activities?"
                ),
                hint="For example eating, washing, walking or going shopping",
                choices={
                    "alot": "Yes, a lot",
                    "little": "Yes, a little",
                    "not": "Not at all",
                    "skip": "Prefer not to say",
                },
                required=False,
                heading="l",
            ),
            predicates={"permission": "yes", "health": "yes"},
        ),
        forms.Question(
            ds.Radios(
                name="ethnic-group",
                label="What is your ethnic group?",
                choices={
                    "white": "White",
                    "mixed": "Mixed or multiple ethnic groups",
                    "asian": "Asian or Asian British",
                    "black": "Black, African, Caribbean or Black British",
                    "other": "Other ethnic group",
                    "skip": "Prefer not to say",
                },
                heading="l",
            ),
            predicates={"permission": "yes"},
        ),
        forms.Question(
            ds.Radios(
                name="white",
                label="Which of the following best describes your White background?",
                choices={
                    "british": "English, Welsh, Scottish, Northern Irish or British",
                    "irish": "Irish",
                    "gypsy": "Gypsy or Irish Traveller",
                    "other": "Any other White background",
                    "skip": "Prefer not to say",
                },
                required=False,
                heading="l",
            ),
            predicates={"permission": "yes", "ethnic-group": "white"},
        ),
        forms.Question(
            ds.Radios(
                name="mixed",
                label=(
                    "Which of the following best describes your "
                    "multiple ethnic group background?"
                ),
                choices={
                    "caribbean": "White and Black Caribbean",
                    "african": "White and Black African",
                    "asian": "White and Asian",
                    "mixed": "Any other mixed or multiple ethnic background",
                    "skip": "Prefer not to say",
                },
                required=False,
                heading="l",
            ),
            predicates={"permission": "yes", "ethnic-group": "mixed"},
        ),
        forms.Question(
            ds.Radios(
                name="asian",
                label=(
                    "Which of the following best describes your "
                    "Asian or Asian British background?"
                ),
                choices={
                    "indian": "Indian",
                    "pakistani": "Pakistani",
                    "bangladeshi": "Bangladeshi",
                    "chinese": "Chinese",
                    "other": "Any other Asian background",
                    "skip": "Prefer not to say",
                },
                required=False,
                heading="l",
            ),
            predicates={"permission": "yes", "ethnic-group": "asian"},
        ),
        forms.Question(
            ds.Radios(
                name="black",
                label=(
                    "Which of the following best describes your Black, African, "
                    "Caribbean or Black British background?"
                ),
                choices={
                    "african": "African",
                    "caribbean": "Caribbean",
                    "other": "Any other Black, African or Caribbean background",
                    "skip": "Prefer not to say",
                },
                required=False,
                heading="l",
            ),
            predicates={"permission": "yes", "ethnic-group": "black"},
        ),
        forms.Question(
            ds.Radios(
                name="other",
                label="Which of the following best describes your background?",
                choices={
                    "arab": "Arab",
                    "other": "Any other ethnic group",
                    "skip": "Prefer not to say",
                },
                required=False,
                heading="l",
            ),
            predicates={"permission": "yes", "ethnic-group": "other"},
        ),
        forms.Question(
            ds.Radios(
                name="marital-status",
                label="What is your legal marital or registered civil partnership status?",
                choices={
                    "never": "Never married and never registered in a civil partnership",
                    "married": "Married",
                    "civil-partnership": "In a registered civil partnership",
                    "separated-married": "Separated, but still legally married",
                    "separated-civil-partnership": "Separated, but still legally in a civil partnership",
                    "divorced": "Divorced",
                    "dissolved": "Formerly in a civil partnership which is now legally dissolved",
                    "widowed": "Widowed",
                    "surviving-partner": "Surviving partner from a registered civil partnership",
                    "skip": "Prefer not to say",
                },
                heading="l",
            ),
            predicates={"permission": "yes"},
        ),
        forms.Question(
            ds.Radios(
                name="religion",
                label="What is your religion?",
                choices={
                    "no": "No religion",
                    "christian": "Christian",
                    "budhhist": "Buddhist",
                    "hindu": "Hindu",
                    "jewish": "Jewish",
                    "muslim": "Muslim",
                    "sikh": "Sikh",
                    "other": "Any other religion",
                    "skip": "Prefer not to say",
                },
                heading="l",
            ),
            predicates={"permission": "yes"},
        ),
        forms.Question(
            ds.Fieldset(
                ds.Radios(
                    name="sex",
                    label="What is your sex?",
                    choices={
                        "female": "Female",
                        "male": "Male",
                        "skip": "Prefer not to say",
                    },
                    heading="m",
                ),
                ds.Radios(
                    name="gender",
                    label=(
                        "Is the gender you identify with the same as "
                        "your sex registered at birth?"
                    ),
                    choices={"yes": "Yes", "no": "No", "skip": "Prefer not to say"},
                    heading="m",
                ),
                legend="Sex and gender identity",
                name="sex-and-gender",
            ),
            predicates={"permission": "yes"},
        ),
        forms.Question(
            ds.Radios(
                name="sexual-orientation",
                label="Which of the following best describes your sexual orientation?",
                choices={
                    "straight": "Heterosexual or straight",
                    "gay-or-lesbian": "Gay or lesbian",
                    "bisexual": "Bisexual",
                    "other": "Other",
                    "skip": "Prefer not to say",
                },
                heading="l",
            ),
            predicates={"permission": "yes"},
            cta="Submit",
        ),
        backends=[forms.DBBackend(db=fast.db)],
        success_url="/",
        data=data,
        step=step,
    )


@fast.page
def cookies():
    """
    This returns the standard GDS cookies page. E.g. -
    https://design-system.service.gov.uk/patterns/cookies-page/full-page/

    It is preset to Essential cookies but if you are using more cookies,
    you can easily add them as follows -

    ds.Cookies(
        ds.H2("Analytical cookies")
        ds.P("...")
        ds.Table(...)
    )

    (See the definition of Cookies)
    """
    return ds.Cookies()


@fast.page
def phase():
    """
    @page does not have to return a full Page. It could also
    return snippets of html.

    E.g.This returns a phase banner snippet that is inserted
    into all the pages of the service using htmx. All you need
    is the following -

        fh.Div(hx_get="/phase", hx_trigger="load"),

    This is useful for several reasons but the primary one is
    that if your service goes Beta, you only need to update this
    and all of the pages on your service will reflect the change.
    """
    return ds.PhaseBanner(
        ds.Span(
            "This is a new service. Help us improve it and ",
            ds.A("give your feedback.", href="/forms/feedback"),
        ),
        phase="Alpha",
    )


@fast.page
def footer():
    """
    This is the footer that will be rendered on every page in your
    service. You can change it once, here and it will change the
    footer on every single page.
    """
    return ds.Footer(("Cookies", "/cookies"))


# Serves the app
fh.serve(app="fast")



def test_home_get(client):
    response = client.get("/")
    assert response.status_code == 200


def test_cookie_banner_get(client):
    response = client.get("/cookie-banner")
    assert response.status_code == 200
    assert 'id="cookie-banner"' in response.text
    assert list(response.cookies.items()) == []

def test_cookie_banner_post(client):
    response = client.post(
        "/cookie-banner",
        data={"cookies[additional]": "hide"},
    )
    assert response.status_code == 200
    assert 'id="cookie-banner"' not in response.text
    assert list(response.cookies.items()) == [
        ("cookie_policy", 'hide')
    ]

def test_cookies_get(client):
    response = client.get("/cookies")
    assert response.status_code == 200
    assert "session_cookie" in response.text
    assert "cookie_policy" in response.text


def test_phase_get(client):
    response = client.get("/phase")
    assert response.status_code == 200
    assert "Alpha" in response.text


def test_notification_get(client):
    client.get("/")
    response = client.get("/notifications")
    assert response.status_code == 200
    assert "Important" in response.text


def test_not_found_get(client):
    response = client.get("/foobar")
    assert response.status_code == 404
    assert "Page not found" in response.text


def test_problem_with_service_get(error_client):
    response = error_client.get("/error")
    assert response.status_code == 500
    assert "Sorry, there is a problem with this service" in response.text


def test_asset_get(client):
    response = client.get("/govuk-frontend-5.11.1.min.css")
    assert response.status_code == 200


def test_page_decorator(fast, client):
    @fast.page
    def test1():
        return
    @fast.page("/foo")
    def test2():
        return
    routes = {r.path for r in fast.routes}
    assert "/test1" in routes
    assert "/foo" in routes
    assert client.get("test1").status_code == 200
    assert client.get("foo").status_code == 200
    assert client.get("bar").status_code == 404


def test_form_decorator(fast, client):
    @fast.form
    def test1(data=None):
        return
    @fast.form("foo")
    def test2(data=None):
        return
    @fast.form()
    def test3(data=None):
        return
    assert "test1" in fast.forms
    assert "foo" in fast.forms
    assert "test3" in fast.forms
    assert client.get("forms/test1").status_code == 200
    assert client.get("forms/foo").status_code == 200
    assert client.get("forms/test3").status_code == 200
    assert client.get("bar").status_code == 404


def test_wizard_decorator(fast, client):
    @fast.wizard
    def test1(step=0, data=None):
        return
    @fast.wizard("foo")
    def test2(step=0, data=None):
        return
    @fast.wizard()
    def test3(step=0, data=None):
        return
    assert "test1" in fast.wizards
    assert "foo" in fast.wizards
    assert "test3" in fast.wizards
    assert client.get("wizards/test1", follow_redirects=True).status_code == 200
    assert client.get("wizards/foo", follow_redirects=True).status_code == 200
    assert client.get("wizards/test3", follow_redirects=True).status_code == 200
    assert client.get("wizards/foo/not-a-step", follow_redirects=True).status_code == 404
    assert client.get("bar", follow_redirects=True).status_code == 404



"""
GOV.UK typography components - paragraphs, headings and lists etc.
"""

import fasthtml.common as fh


def A(
    text,
    href: str = "#",
    visited: bool = False,
    inverse=False,
    newtab=False,
    **kwargs,
) -> fh.FT:
    """
    `GOV.UK Link`_ component. Links are blue and underlined by default.

    Examples:

        >>> a = ds.A("test")
        >>> str(a)
        '<a href="#" class="govuk-link">test</a>'
        >>> a = ds.A("test", href="/test")
        >>> str(a)
        '<a href="/test" class="govuk-link">test</a>'

    Args:
        text: The text to display in the link.
        href (str, optional): The URL the link points to. Defaults to "#".
        visited (bool, optional): If True, applies a visited style. Defaults to False.
        inverse (bool, optional): If True, applies an inverse style. Defaults to False.
        newtab (bool, optional): If True, opens the link in a new tab. Defaults to False.
        **kwargs: Additional keyword arguments.

    Returns:
        FT: A FastHTML link component.

    .. _GOV.UK Link: https://design-system.service.gov.uk/styles/links/
    """
    visited_cls = " govuk-link--visited" if visited else ""
    inverse_cls = " govuk-link--inverse" if inverse else ""
    cls = f"govuk-link{visited_cls}{inverse_cls}"
    newtab_attr = {"target": "_blank", "rel": "noopener noreferrer"} if newtab else {}
    return fh.A(text, href=href, cls=cls, **newtab_attr, **kwargs)


def H1(text, size="l", caption="", **kwargs) -> fh.FT:
    """
    `GOV.UK H1`_ component.

    Examples:

        >>> h1 = ds.H1("Test")
        >>> str(h1)
        '<h1 class="govuk-heading-l">Test</h1>'
        >>> h1 = ds.H1("Test", size="xl")
        >>> str(h1)
        '<h1 class="govuk-heading-xl">Test</h1>'
        >>> h1 = ds.H1("Test", caption="test caption")
        >>> str(h1)
        '<h1 class="govuk-heading-l">Test<span class="govuk-caption-l">test caption</span></h1>'

    Args:
        text: The text to display in the header.
        size (str, optional): The size of the header. Defaults to "l".
        caption (str, optional): Caption to go with the heading. Defaults to "".
        **kwargs: Additional keyword arguments.

    Returns:
        FT: A FastHTML H1 component.

    .. _GOV.UK H1: https://design-system.service.gov.uk/styles/headings/
    """
    return fh.H1(
        text,
        fh.Span(caption, cls=f"govuk-caption-{size}") if caption else "",
        cls=f"govuk-heading-{size}",
        **kwargs,
    )


def H2(text, size="m", caption="", **kwargs) -> fh.FT:
    """
    `GOV.UK H2`_ component.

    Examples:

        >>> h2 = ds.H2("Test")
        >>> str(h2)
        '<h2 class="govuk-heading-l">Test</h2>'
        >>> h2 = ds.H2("Test", size="xl")
        >>> str(h2)
        '<h2 class="govuk-heading-xl">Test</h2>'
        >>> h2 = ds.H2("Test", caption="test caption")
        >>> str(h2)
        '<h2 class="govuk-heading-l">Test<span class="govuk-caption-l">test caption</span></h2>'

    Args:
        text: The text to display in the header.
        size (str, optional): The size of the header. Defaults to "l".
        caption (str, optional): Caption to go with the heading. Defaults to "".
        **kwargs: Additional keyword arguments.

    Returns:
        FT: A FastHTML H2 component.

    .. _GOV.UK H2: https://design-system.service.gov.uk/styles/headings/
    """
    return fh.H2(
        text,
        fh.Span(caption, cls=f"govuk-caption-{size}") if caption else "",
        cls=f"govuk-heading-{size}",
        **kwargs,
    )


def H3(text, size="s", caption="", **kwargs) -> fh.FT:
    """
    `GOV.UK H2`_ component.

    Examples:

        >>> h2 = ds.H2("Test")
        >>> str(h2)
        '<h2 class="govuk-heading-l">Test</h2>'
        >>> h2 = ds.H2("Test", size="xl")
        >>> str(h2)
        '<h2 class="govuk-heading-xl">Test</h2>'
        >>> h2 = ds.H2("Test", caption="test caption")
        >>> str(h2)
        '<h2 class="govuk-heading-l">Test<span class="govuk-caption-l">test caption</span></h2>'

    Args:
        text: The text to display in the header.
        size (str, optional): The size of the header. Defaults to "l".
        caption (str, optional): Caption to go with the heading. Defaults to "".
        **kwargs: Additional keyword arguments.

    Returns:
        FT: A FastHTML H2 component.

    .. _GOV.UK H2: https://design-system.service.gov.uk/styles/headings/
    """
    return fh.H3(
        text,
        fh.Span(caption, cls=f"govuk-caption-{size}") if caption else "",
        cls=f"govuk-heading-{size}",
        **kwargs,
    )


def P(*content, lead=False, small=False, **kwargs) -> fh.FT:
    """
    `GOV.UK Paragraph`_ component. The default paragraph font size is 19px.

    Examples:

        >>> p = ds.P("test")
        >>> str(p)
        '<p class="govuk-body">test</p>'
        >>> p = ds.P("test", lead=True)
        >>> str(p)
        '<p class="govuk-body-l">test</p>'
        >>> p = ds.P("test", small=True)
        >>> str(p)
        '<p class="govuk-body-s">test</p>'

    Args:
        *content: The list of content to display in the paragraph.
        lead (bool, optional): If True, applies a lead style. Defaults to False.
        small (bool, optional): If True, applies a small style. Defaults to False.
        **kwargs: Additional keyword arguments.

    Returns:
        FT: A FastHTML paragraph component.

    .. _GOV.UK Paragraph: https://design-system.service.gov.uk/styles/paragraphs/
    """
    if lead and small:
        raise ValueError("Cannot set both lead and small to True.")
    cls_suffix = "-l" if lead else "-s" if small else ""

    return fh.P(*content, cls=f"govuk-body{cls_suffix}", **kwargs)


def Li(*args, **kwargs) -> fh.FT:
    """
    `GOV.UK list item` component. Used to pass-in list items into :py:meth:`Ul`.

    Args:
        *args: Items to include in the list.
        **kwargs: Additional attributes for the list.

    Returns:
        FT: A FastHTML list item component.

    .. _GOV.UK list item: https://design-system.service.gov.uk/styles/lists/
    """
    return fh.Li(*args, **kwargs)


def Ul(*args, bullet=False, numbered=False, spaced=False, **kwargs) -> fh.FT:
    """
    `GOV.UK unordered list`_ component. Use lists to make blocks of text easier to read, and to
    break information into manageable chunks.

    Examples:

        >>> li = ds.Li("test")
        >>> ul = ds.Ul(li)
        >>> str(ul)
        '<ul class="govuk-list"><li>test</li></ul>'
        >>> ul = ds.Ul(li, bullet=True)
        >>> str(ul)
        '<ul class="govuk-list govuk-list--bullet"><li>test</li></ul>'
        >>> ul = ds.Ul(li, numbered=True)
        >>> str(ul)
        '<ul class="govuk-list govuk-list--number"><li>test</li></ul>'

    Args:
        *args: Items to include in the list.
        bullet (bool, optional): If True, applies a bullet style. Defaults to False.
        numbered (bool, optional): If True, applies a numbered style. Defaults to False.
        spaced (bool, optional): If True, applies a spaced style. Defaults to False.
        **kwargs: Additional attributes for the list.

    Returns:
        FT: A FastHTML unordered list component.

    .. _GOV.UK unordered list: https://design-system.service.gov.uk/styles/lists/
    """
    if bullet and numbered:
        raise ValueError("Cannot set both bullet and numbered to True.")
    if bullet:
        cls = "govuk-list govuk-list--bullet"
    elif numbered:
        cls = "govuk-list govuk-list--number"
    else:
        cls = "govuk-list"
    if spaced:
        cls += " govuk-list--spaced"
    return fh.Ul(*args, cls=cls, **kwargs)



import fasthtml.common as fh
import pytest

import fast_gov_uk.design_system as ds


@pytest.mark.parametrize(
    "kwargs, expected",
    (
        (
            {"text": "Test"},
            '<div class="govuk-inset-text">Test</div>',
        ),
    ),
)
def test_inset(kwargs, expected, html):
    """Test Inset text with various parameters.
    Args:
        kwargs (dict): The arguments to pass to Inset.
        expected (str): The expected HTML output.
    """
    text = ds.Inset(**kwargs)
    assert html(text) == html(expected)


@pytest.mark.parametrize(
    "content, expected",
    (
        (
            ["Test Content"],
            (
                '<details class="govuk-details">'
                    '<summary class="govuk-details__summary">'
                    '<span class="govuk-details__summary-text">Test</span>'
                    '</summary>'
                    '<div class="govuk-details__text">'
                        'Test Content'
                    '</div>'
                '</details>'
            ),
        ),
        (
            [ds.A("Test Link")],
            (
                '<details class="govuk-details">'
                    '<summary class="govuk-details__summary">'
                        '<span class="govuk-details__summary-text">Test</span>'
                    '</summary>'
                    '<div class="govuk-details__text">'
                        '<a href="#" class="govuk-link">Test Link</a>'
                    '</div>'
                '</details>'
            ),
        ),
        (
            [ds.A("Test Link"), ds.P("Test para")],
            (
                '<details class="govuk-details">'
                    '<summary class="govuk-details__summary">'
                        '<span class="govuk-details__summary-text">Test</span>'
                    '</summary>'
                    '<div class="govuk-details__text">'
                        '<a href="#" class="govuk-link">Test Link</a>'
                        '<p class="govuk-body">Test para</p>'
                    '</div>'
                '</details>'
            ),
        ),
    ),
)
def test_detail(content, expected, html):
    """Test Detail with various parameters.
    Args:
        args (dict): The arguments to pass to Detail.
        expected (str): The expected HTML output.
    """
    detail = ds.Detail("Test", *content)
    assert html(detail) == html(expected)


@pytest.mark.parametrize(
    "kwargs, expected",
    (
        (
            {"content": [ds.P("Test Content")]},
            (
                '<div class="govuk-panel govuk-panel--confirmation">'
                    '<div class="govuk-panel__body">'
                        '<p class="govuk-body">Test Content</p>'
                    '</div>'
                '</div>'
            ),
        ),
        (
            {"title": "Test", "content": [ds.P("Test Content")]},
            (
                '<div class="govuk-panel govuk-panel--confirmation">'
                    '<h1 class="govuk-panel__title">Test</h1>'
                    '<div class="govuk-panel__body">'
                        '<p class="govuk-body">Test Content</p>'
                    '</div>'
                '</div>'
            ),
        ),
        (
            {"content": [ds.A("Test Link")]},
            (
                '<div class="govuk-panel govuk-panel--confirmation">'
                    '<div class="govuk-panel__body">'
                        '<a href="#" class="govuk-link">Test Link</a>'
                    '</div>'
                '</div>'
            ),
        ),
        (
            {"content": [ds.A("Test Link"), ds.P("Test Content")]},
            (
                '<div class="govuk-panel govuk-panel--confirmation">'
                    '<div class="govuk-panel__body">'
                        '<a href="#" class="govuk-link">Test Link</a>'
                        '<p class="govuk-body">Test Content</p>'
                    '</div>'
                '</div>'
            ),
        ),
    ),
)
def test_panel(kwargs, expected, html):
    """Test Panel with various parameters.
    Args:
        kwargs (dict): The arguments to pass to Panel.
        expected (str): The expected HTML output.
    """
    content = kwargs.pop("content")
    panel = ds.Panel(*content, **kwargs)
    assert html(panel) == html(expected)


@pytest.mark.parametrize(
    "kwargs, expected",
    (
        (
            {"text": "Test"},
            '<strong class="govuk-tag">Test</strong>',
        ),
        (
            {"text": "Test", "color": "blue"},
            '<strong class="govuk-tag govuk-tag--blue">Test</strong>',
        ),
        (
            {"text": "Test", "color": "grey"},
            '<strong class="govuk-tag govuk-tag--grey">Test</strong>',
        ),
        (
            {"text": "Test", "color": "red"},
            '<strong class="govuk-tag govuk-tag--red">Test</strong>',
        ),
        (
            {"text": "Test", "color": "green"},
            '<strong class="govuk-tag govuk-tag--green">Test</strong>',
        ),
        (
            {"text": "Test", "color": "yellow"},
            '<strong class="govuk-tag govuk-tag--yellow">Test</strong>',
        ),
    ),
)
def test_tag(kwargs, expected, html):
    """Test Panel with various parameters.
    Args:
        kwargs (dict): The arguments to pass to Tag.
        expected (str): The expected HTML output.
    """
    tag = ds.Tag(**kwargs)
    assert html(tag) == html(expected)


@pytest.mark.parametrize(
    "kwargs, expected",
    (
        (
            {"content": [ds.P("You can be fined up to £5,000 if you do not register.")]},
            (
                '<div class="govuk-warning-text">'
                    '<span aria-hidden="true" class="govuk-warning-text__icon">!</span>'
                        '<strong class="govuk-warning-text__text">'
                            '<span class="govuk-visually-hidden">Warning</span>'
                            '<p class="govuk-body">You can be fined up to £5,000 if you do not register.</p>'
                        "</strong>"
                "</div>"
            ),
        ),
    ),
)
def test_warning(kwargs, expected, html):
    """Test Warning with various parameters.
    Args:
        kwargs (dict): The arguments to pass to Warning.
        expected (str): The expected HTML output.
    """
    content = kwargs.pop("content")
    banner = ds.Warning(*content, **kwargs)
    assert html(banner) == html(expected)


@pytest.mark.parametrize(
    "kwargs, expected",
    (
        (
            {"title": "Important", "content": ["Test Content"]},
            (
                '<div role="alert" aria-labelledby="govuk-notification-banner-title" data-module="govuk-notification-banner" class="govuk-notification-banner">'
                    '<div class="govuk-notification-banner__header">'
                        '<h2 id="govuk-notification-banner-title" class="govuk-notification-banner__title">Important</h2>'
                    "</div>"
                    '<div class="govuk-notification-banner__content">'
                        "Test Content"
                    "</div>"
                "</div>"
            ),
        ),
        (
            {"title": "Success", "content": ["Test Content"], "success": True},
            (
                '<div role="alert" aria-labelledby="govuk-notification-banner-title" data-module="govuk-notification-banner" class="govuk-notification-banner govuk-notification-banner--success">'
                    '<div class="govuk-notification-banner__header">'
                        '<h2 id="govuk-notification-banner-title" class="govuk-notification-banner__title">Success</h2>'
                    "</div>"
                    '<div class="govuk-notification-banner__content">'
                        "Test Content"
                    "</div>"
                "</div>"
            ),
        ),
        (
            {"title": "Important", "content": [ds.A("Test Link")]},
            (
                '<div role="alert" aria-labelledby="govuk-notification-banner-title" data-module="govuk-notification-banner" class="govuk-notification-banner">'
                    '<div class="govuk-notification-banner__header">'
                        '<h2 id="govuk-notification-banner-title" class="govuk-notification-banner__title">Important</h2>'
                    "</div>"
                    '<div class="govuk-notification-banner__content">'
                        '<a href="#" class="govuk-link">Test Link</a>'
                    "</div>"
                "</div>"
            ),
        ),
    ),
)
def test_notification(kwargs, expected, html):
    """Test Notification with various parameters.
    Args:
        kwargs (dict): The arguments to pass to Notification.
        expected (str): The expected HTML output.
    """
    content = kwargs.pop("content")
    notification = ds.Notification(*content, **kwargs)
    assert html(notification) == html(expected)


@pytest.mark.parametrize(
    "sections, expected",
    (
        (
            [{"heading": "Test 1", "content": "Test Content"}],
            (
                "<div>"
                    '<div data-module="govuk-accordion" class="govuk-accordion">'
                        '<div class="govuk-accordion__section">'
                            '<div class="govuk-accordion__section-header">'
                                '<h2 class="govuk-accordion__section-heading">'
                                    '<span id="accordion-heading-1" class="govuk-accordion__section-button">'
                                        "Test 1"
                                    "</span>"
                                "</h2>"
                            "</div>"
                            '<div id="accordion-content-1" class="govuk-accordion__section-content">'
                                "Test Content"
                            "</div>"
                        "</div>"
                    "</div>"
                "</div>"
            ),
        ),
        (
            [
                {"heading": "Test 1", "content": "Test Content 1"},
                {"heading": "Test 2", "content": "Test Content 2"},
            ],
            (
                "<div>"
                    '<div data-module="govuk-accordion" class="govuk-accordion">'
                        '<div class="govuk-accordion__section">'
                            '<div class="govuk-accordion__section-header">'
                                '<h2 class="govuk-accordion__section-heading">'
                                    '<span id="accordion-heading-1" class="govuk-accordion__section-button">'
                                        "Test 1"
                                    "</span>"
                                "</h2>"
                            "</div>"
                            '<div id="accordion-content-1" class="govuk-accordion__section-content">'
                                "Test Content 1"
                            "</div>"
                        "</div>"
                        '<div class="govuk-accordion__section">'
                            '<div class="govuk-accordion__section-header">'
                                '<h2 class="govuk-accordion__section-heading">'
                                    '<span id="accordion-heading-2" class="govuk-accordion__section-button">'
                                        "Test 2"
                                    "</span>"
                                "</h2>"
                            "</div>"
                            '<div id="accordion-content-2" class="govuk-accordion__section-content">'
                                "Test Content 2"
                            "</div>"
                        "</div>"
                    "</div>"
                "</div>"
            ),
        ),
    ),
)
def test_accordion(sections, expected, html):
    """Test accordion with various parameters.
    Args:
        sections (list): The sections to pass to accordion.
        expected (str): The expected HTML output.
    """
    accordion = ds.Div(ds.Accordion(*sections))
    assert html(accordion) == html(expected)


@pytest.mark.parametrize(
    "panels, expected",
    (
        (
            [
                {"heading": "Test 1", "content": "Test Content 1"},
                {"heading": "Test 2", "content": "Test Content 2"},
            ],
            (
                '<div data-module="govuk-tabs" class="govuk-tabs">'
                    '<h2 class="govuk-tabs__title"></h2>'
                    '<ul class="govuk-tabs__list">'
                        '<li class="govuk-tabs__list-item govuk-tabs__list-item--selected">'
                            '<a href="#test-1" class="govuk-tabs__tab">Test 1</a>'
                        "</li>"
                        '<li class="govuk-tabs__list-item">'
                            '<a href="#test-2" class="govuk-tabs__tab">Test 2</a>'
                        "</li>"
                    "</ul>"
                    '<div id="test-1" class="govuk-tabs__panel">'
                        '<h2 class="govuk-heading-l">Test 1</h2>'
                        "Test Content 1"
                    "</div>"
                    '<div id="test-2" class="govuk-tabs__panel govuk-tabs__panel--hidden">'
                        '<h2 class="govuk-heading-l">Test 2</h2>'
                        "Test Content 2"
                    "</div>"
                "</div>"
            ),
        ),
    ),
)
def test_tabs(panels, expected, html):
    """Test Tab with various parameters.
    Args:
        panels (list): The panels to pass to Tab.
        expected (str): The expected HTML output.
    """
    tab = ds.Tab(*panels)
    assert html(tab) == html(expected)


def test_errorsummary(html):
    """Test ErrorSummary with various parameters."""
    summary = ds.ErrorSummary(
        "Test Legend",
        fh.A("Test 1", href="/test1"),
        fh.A("Test 2", href="/test2"),
    )
    expected = (
        '<div data-module="govuk-error-summary" class="govuk-error-summary">'
            '<div role="alert">'
                '<h2 class="govuk-error-summary__title">Test Legend</h2>'
                '<div class="govuk-error-summary__body">'
                    '<ul class="govuk-list govuk-error-summary__list">'
                        '<li><a href="/test1">Test 1</a></li>'
                        '<li><a href="/test2">Test 2</a></li>'
                    "</ul>"
                "</div>"
            "</div>"
        "</div>"
    )
    assert html(summary) == html(expected)


def test_table(html):
    """
    Test Table with various parameters.
    """
    table = ds.Table(
        [
            {"Fruit": "Apple", "Price": "£0.25"},
            {"Fruit": "Orange", "Price": "£0.5"},
            {"Fruit": "Banana", "Price": "£0.1"},
        ],
        caption="Test",
        header_cols=["Fruit"],
    )
    expected = (
        '<table class="govuk-table">'
            '<caption class="govuk-table__caption govuk-table__caption--m">Test</caption>'
            '<thead class="govuk-table__head">'
                '<tr class="govuk-table__row">'
                    '<th scope="col" class="govuk-table__header">Fruit</th>'
                    '<th scope="col" class="govuk-table__header">Price</th>'
                "</tr>"
            "</thead>"
            '<tbody class="govuk-table__body">'
                '<tr class="govuk-table__row">'
                    '<th class="govuk-table__header" scope="row">Apple</th>'
                    '<td class="govuk-table__cell">£0.25</td>'
                "</tr>"
                '<tr class="govuk-table__row">'
                    '<th class="govuk-table__header" scope="row">Orange</th>'
                    '<td class="govuk-table__cell">£0.5</td>'
                "</tr>"
                '<tr class="govuk-table__row">'
                    '<th class="govuk-table__header" scope="row">Banana</th>'
                    '<td class="govuk-table__cell">£0.1</td>'
                "</tr>"
            "</tbody>"
        "</table>"
    )
    assert html(table) == html(expected)


@pytest.mark.parametrize(
    "kwargs, expected",
    (
        (
            {"label": "test", "href": "/test"},
            (
                '<li class="govuk-task-list__item govuk-task-list__item--with-link">'
                    '<div class="govuk-task-list__name-and-hint">'
                        '<a href="/test" aria-describedby="test-status" class="govuk-link govuk-task-list__link">test</a>'
                    "</div>"
                    '<div id="test-status" class="govuk-task-list__status">'
                        '<strong class="govuk-tag govuk-tag--blue">Incomplete</strong>'
                    "</div>"
                "</li>"
            ),
        ),
        (
            {"label": "test", "href": "/test", "completed": True},
            (
                '<li class="govuk-task-list__item govuk-task-list__item--with-link">'
                    '<div class="govuk-task-list__name-and-hint">'
                        '<a href="/test" aria-describedby="test-status" class="govuk-link govuk-task-list__link">test</a>'
                    "</div>"
                    '<div id="test-status" class="govuk-task-list__status">Completed</div>'
                "</li>"
            ),
        ),
        (
            {"label": "test", "href": "/test", "completed": True, "hint": "Test Hint"},
            (
                '<li class="govuk-task-list__item govuk-task-list__item--with-link">'
                    '<div class="govuk-task-list__name-and-hint">'
                        '<a href="/test" aria-describedby="test-hint test-status" class="govuk-link govuk-task-list__link">test</a>'
                        '<div id="test-hint" class="govuk-task-list__hint">Test Hint</div>'
                    "</div>"
                    '<div id="test-status" class="govuk-task-list__status">Completed</div>'
                "</li>"
            ),
        ),
    ),
)
def test_task(kwargs, expected, html):
    """Test Task with various parameters.
    Args:
        kwargs (dict): The arguments to pass to Task.
        expected (str): The expected HTML output.
    """
    task = ds.Task(**kwargs)
    assert html(task) == html(expected)


@pytest.mark.parametrize(
    "args, expected",
    (
        (
            [ds.Task("Test Label 1", "/test1"), ds.Task("Test Label 2", "/test2")],
            (
                '<ul class="govuk-task-list">'
                    '<li class="govuk-task-list__item govuk-task-list__item--with-link">'
                        '<div class="govuk-task-list__name-and-hint">'
                            '<a href="/test1" aria-describedby="test-label-1-status" class="govuk-link govuk-task-list__link">Test Label 1</a>'
                        "</div>"
                        '<div id="test-label-1-status" class="govuk-task-list__status">'
                            '<strong class="govuk-tag govuk-tag--blue">Incomplete</strong>'
                        "</div>"
                    "</li>"
                    '<li class="govuk-task-list__item govuk-task-list__item--with-link">'
                        '<div class="govuk-task-list__name-and-hint">'
                            '<a href="/test2" aria-describedby="test-label-2-status" class="govuk-link govuk-task-list__link">Test Label 2</a>'
                        "</div>"
                        '<div id="test-label-2-status" class="govuk-task-list__status">'
                            '<strong class="govuk-tag govuk-tag--blue">Incomplete</strong>'
                        "</div>"
                    "</li>"
                "</ul>"
            ),
        ),
    ),
)
def test_tasklist(args, expected, html):
    """Test TaskList with various parameters.
    Args:
        args (list): The arguments to pass to TaskList.
        expected (str): The expected HTML output.
    """
    tl = ds.TaskList(*args)
    assert html(tl) == html(expected)


@pytest.mark.parametrize(
    "args, expected",
    (
        (
            [
                ds.SummaryItem("Name", "John Doe"),
                ds.SummaryItem("DOB", "", ds.A("Test Label 1", "/test1")),
                ds.SummaryItem(
                    "Email",
                    "test@test.com",
                    ds.A("Test Label 2", "/test2"), ds.A("Test Label 2", "/test2"),
                ),
            ],
            (
                '<dl class="govuk-summary-list">'
                    '<div class="govuk-summary-list__row govuk-summary-list__row--no-actions">'
                        '<dt class="govuk-summary-list__key">Name</dt>'
                        '<dd class="govuk-summary-list__value">John Doe</dd>'
                    "</div>"
                    '<div class="govuk-summary-list__row">'
                        '<dt class="govuk-summary-list__key">DOB</dt>'
                        '<dd class="govuk-summary-list__value"></dd>'
                        '<dd class="govuk-summary-list__actions">'
                            '<a href="/test1" class="govuk-link">'
                                'Test Label 1 <span class="govuk-visually-hidden">dob</span>'
                            '</a>'
                        "</dd>"
                    "</div>"
                    '<div class="govuk-summary-list__row">'
                        '<dt class="govuk-summary-list__key">Email</dt>'
                        '<dd class="govuk-summary-list__value">test@test.com</dd>'
                        '<dd class="govuk-summary-list__actions">'
                            '<ul class="govuk-summary-list__actions-list">'
                                '<li class="govuk-summary-list__actions-list-item">'
                                    '<a href="/test2" class="govuk-link">'
                                        'Test Label 2 <span class="govuk-visually-hidden">email</span>'
                                    '</a>'
                                "</li>"
                                '<li class="govuk-summary-list__actions-list-item">'
                                    '<a href="/test2" class="govuk-link">'
                                        'Test Label 2 <span class="govuk-visually-hidden">email</span>'
                                    '</a>'
                                "</li>"
                            "</ul>"
                        "</dd>"
                    "</div>"
                "</div>"
            ),
        ),
    ),
)
def test_summary_list(args, expected, html):
    """Test SummaryList with various parameters.
    Args:
        args (list): The arguments to pass to SummaryList.
        expected (str): The expected HTML output.
    """
    summary = ds.SummaryList(*args)
    assert html(summary) == html(expected)


@pytest.mark.parametrize(
    "args, expected",
    (
        (
            [
                ds.SummaryItem(
                    "Email",
                    "test@test.com",
                    ds.A("Test Label 1", "/test1"),
                    ds.A("Test Label 2", "/test2"),
                ),
            ],
            (
                '<div class="govuk-summary-card">'
                    '<div class="govuk-summary-card__title-wrapper">'
                        '<h2 class="govuk-summary-card__title">Test</h2>'
                        '<ul class="govuk-summary-card__actions">'
                            '<li class="govuk-summary-card__action">'
                                '<a href="/test1" class="govuk-link">'
                                    'Test Action 1 <span class="govuk-visually-hidden">(Test)</span>'
                                '</a>'
                            "</li>"
                        "</ul>"
                    "</div>"
                    '<div class="govuk-summary-card__content">'
                        '<dl class="govuk-summary-list">'
                            '<div class="govuk-summary-list__row">'
                                '<dt class="govuk-summary-list__key">Email</dt>'
                                '<dd class="govuk-summary-list__value">test@test.com</dd>'
                                '<dd class="govuk-summary-list__actions">'
                                    '<ul class="govuk-summary-list__actions-list">'
                                        '<li class="govuk-summary-list__actions-list-item">'
                                            '<a href="/test1" class="govuk-link">'
                                                'Test Label 1 <span class="govuk-visually-hidden">email</span>'
                                            '</a>'
                                        "</li>"
                                        '<li class="govuk-summary-list__actions-list-item">'
                                            '<a href="/test2" class="govuk-link">'
                                                'Test Label 2 <span class="govuk-visually-hidden">email</span>'
                                            '</a>'
                                        "</li>"
                                    "</ul>"
                                "</dd>"
                            "</div>"
                        "</div>"
                    "</div>"
                "</div>"
            ),
        ),
    ),
)
def test_summary_card(args, expected, html):
    """Test SummaryCard with various parameters.
    Args:
        args (list): The arguments to pass to SummaryCard.
        expected (str): The expected HTML output.
    """
    summary = ds.SummaryCard(
        title="Test",
        summary_list=ds.SummaryList(*args),
        actions=[ds.A("Test Action 1", "/test1")],
    )
    assert html(summary) == html(expected)


@pytest.mark.parametrize("component", (
    ds.Inset("test", hx_test="foo"),
    ds.Detail("test", hx_test="foo"),
    ds.Panel(hx_test="foo"),
    ds.Tag("test", hx_test="foo"),
    ds.Warning(hx_test="foo"),
    ds.Notification(hx_test="foo"),
    ds.Accordion(hx_test="foo"),
    ds.Tab(hx_test="foo"),
    ds.ErrorSummary("test", hx_test="foo"),
    ds.Table(data=[{}], hx_test="foo"),
    ds.Task("test", "/", hx_test="foo"),
    ds.TaskList(hx_test="foo"),
    ds.SummaryItem("test", "test", hx_test="foo"),
    ds.SummaryList(hx_test="foo"),
    ds.SummaryCard("test", ds.P(""), hx_test="foo"),
))
def test_html_attribute(component, html):
    """
    Test that passes an html attribute to components - that is not
    explicitly handled but should be passed through to the
    underlying FT.
    """
    assert 'hx-test="foo"' in html(component)



import fasthtml.common as fh

import fast_gov_uk.design_system as ds

DEMO_PAGE = (
    ds.H1("H1 Heading"),
    ds.H2("H2 Heading"),
    ds.H3("H3 Heading"),
    ds.P("Paragraph"),
    ds.A("Link"),
    ds.Ul(
        ds.Li("Item 1"),
        ds.Li("Item 2"),
        ds.Li("Item 3"),
    ),
    ds.Ul(
        ds.Li("Item 1"),
        ds.Li("Item 2"),
        ds.Li("Item 3"),
        bullet=True,
    ),
    ds.Ul(
        ds.Li("Item 1"),
        ds.Li("Item 2"),
        ds.Li("Item 3"),
        numbered=True,
    ),
    ds.Ul(
        ds.Li("Item 1"),
        ds.Li("Item 2"),
        ds.Li("Item 3"),
        spaced=True,
    ),
    ds.Inset("This is an inset paragraph."),
    ds.Detail("A detail component", ds.P("Here is some content.")),
    ds.Panel(ds.P("Here is some content"), title="Panel component"),
    ds.Tag("A Tag"),
    fh.Nbsp(),
    ds.Tag("Green Tag", color="green"),
    fh.Nbsp(),
    ds.Tag("Red Tag", color="red"),
    fh.Nbsp(),
    ds.Tag("Yellow Tag", color="yellow"),
    fh.Nbsp(),
    ds.Tag("Grey Tag", color="grey"),
    fh.Nbsp(),
    ds.PhaseBanner(ds.P("Phase Banner"), phase="apha"),
    ds.Warning(ds.P("Here is a Warning")),
    ds.Notification(ds.P("Here is some content"), title="A Notification"),
    ds.Notification(ds.P("Here is some content"), title="A Notification", success=True),
    ds.Accordion(
        {"heading": "An accordion Section", "content": "Some content"},
        {"heading": "Another accordion Section", "content": "Some more content"},
    ),
    ds.Tab(
        {"heading": "Tab 1", "content": "Some content"},
        {"heading": "Tab 2", "content": "Some more content"},
    ),
    ds.Breadcrumbs(
        ("A breadcrumb link", "#"),
        ("Another breadcrumb link", "#"),
    ),
    ds.Select(
        name="color",
        choices={"Blue": "blue", "Red": "red", "Green": "green"},
        label="A Select:",
        hint="With a hint",
    ),
    ds.Textarea(
        "detail",
        label="A Textarea",
        hint="With a hint",
        heading="l",
    ),
    ds.PasswordInput("pass", label="A Password input", hint="with a helpful hint"),
    ds.CharacterCount(
        "char",
        label="Character Count",
        hint="with a helpful hint",
        maxchars=10,
        heading="l",
    ),
    ds.ErrorSummary("An Error Summary", ds.A("A link")),
    ds.Table(
        [
            {"Fruit": "Apple", "Price": "£0.5"},
            {"Fruit": "Orange", "Price": "£0.75"},
            {"Fruit": "Banana", "Price": "£0.25"},
        ],
        caption="A Table",
        header_cols=["Fruit"],
    ),
    ds.Button("A Button"),
    fh.Nbsp(),
    ds.Button("Secondary Button", style=ds.ButtonStyle.SECONDARY),
    fh.Nbsp(),
    ds.Button("Warning Button", style=ds.ButtonStyle.WARNING),
    fh.Nbsp(),
    ds.Button("Inverse Button", style=ds.ButtonStyle.INVERSE),
    fh.Nbsp(),
    ds.TextInput(
        "input",
        label="A text input",
        hint="With a hint",
        heading="l",
        prefix="£",
        suffix="Only",
    ),
    ds.Checkboxes(
        name="check",
        checkboxes=[
            ds.Checkbox("check", "1", "Apple", hint="Choose Apple"),
            ds.Checkbox("check", "2", "Orange", hint="Choose Orange"),
            ds.Checkbox("check", "3", "Banana", hint="Choose Banana"),
        ],
        label="Checkboxes",
        hint="With a hint",
        heading="l",
    ),
    ds.Radios(
        name="radio",
        radios=[
            ds.Radio("radio", "1", "Apple", hint="Choose Apple"),
            ds.Radio("radio", "2", "Orange", hint="Choose Orange"),
            ds.Radio("radio", "3", "Banana", hint="Choose Banana"),
        ],
        label="Radios",
        hint="With a hint",
        heading="l",
    ),
    ds.FileUpload(
        "file_upload",
        "File Upload",
        hint="With a hint",
        heading="l",
    ),
    ds.ExitPage(text="Exit Page component"),
    ds.Navigation(
        ds.NavigationLink("A link", "#"),
        ds.NavigationLink("Another link", "#"),
        service_name="Name of the service",
    ),
    ds.Pagination(
        ds.PaginationLink("1", "/test?page=1"),
        ds.PaginationLink("2", "/test?page=2"),
        ds.PaginationLink("3", "/test?page=3", active=True),
        ds.PaginationLink("4", "/test?page=4"),
        ds.PaginationLink("5", "/test?page=5"),
        prev_link="/test?page=2",
        next_link="/test?page=4",
    ),
    ds.DateInput("date_input", "A Date Input", hint="With a hint", heading="l"),
    ds.SummaryList(
        ds.SummaryItem("Name", "John Doe"),
        ds.SummaryItem("Date of Birth", "13/10/2010", ds.A("Change", "/change-dob")),
        ds.SummaryItem(
            "Email",
            "test@test.com",
            ds.A("Add", "/add-email"),
            ds.A("Change", "/change-email"),
        ),
    ),
    ds.SummaryCard(
        "A Summary Card",
        ds.SummaryList(
            ds.SummaryItem("Name", "John Doe"),
            ds.SummaryItem("Date of Birth", "13/10/2010"),
            ds.SummaryItem(
                "Email",
                "test@test.com",
                ds.A("Add", "/add-email"),
                ds.A("Change", "/change-email"),
            ),
        ),
        actions=[ds.A("Delete", "/delete-card")],
    ),
    ds.CookieBanner(
        "Test Service",
        ds.P("Here is a cookie banner"),
        ds.P("We use cookies for analytics."),
    ),
)


def demo():
    return ds.Page(*DEMO_PAGE)



import json
from dataclasses import asdict
from unittest.mock import Mock, patch, call, ANY

import pytest

from .app import session_feedback
from fast_gov_uk import forms


def test_form_get(client):
    response = client.get("/forms/profile")
    assert response.status_code == 200


def test_form_get_404(client):
    response = client.get("/forms/not-profile")
    assert response.status_code == 404


def test_db_form_post_valid(client, db, picture):
    data = {
        "name": "Test",
        "sex": "male",
        "gender": "yes",
        "ethnicity": "mixed",
        "dob": ["10", "10", "2000"],
        "phone": "12345",
        "email": "test@test",
        "comments": "Test",
    }
    response = client.post(
        "/forms/profile",
        data=data,
        files={"picture": picture},
    )
    assert response.status_code == 303
    forms = db.t.forms()
    form = forms[0]
    form_json = asdict(form)
    form_data = form_json["data"]
    form_dict = json.loads(form_data)
    assert form_dict == {
        "name": "Test",
        "sex": "Male",
        "gender": "Yes",
        "ethnicity": "Mixed",
        # This should be a date string
        "dob": "2000-10-10",
        # This should be a path to the file stored in /media
        "picture": "media/picture.png",
        # This should be a number
        "phone": 12345,
        "email": "test@test",
        "comments": "Test",
    }


@pytest.mark.parametrize("errors, expected", (
        (
            # empty name
            {"name": ""},
            ("name", "This field is required.")
        ),
        (
            # empty sex
            {"sex": ""},
            ("sex", "This field is required.")
        ),
        (
            # empty gender
            {"gender": ""},
            ("gender", "This field is required.")
        ),
        (
            # empty ethnicity
            {"ethnicity": ""},
            ("ethnicity", "This field is required.")
        ),
        (
            # empty dob
            {"dob": ["", "", ""]},
            ("dob", "This field is required.")
        ),
        (
            # partially empty dob
            {"dob": ["10", "10", ""]},
            ("dob", "This field is required.")
        ),
        (
            # invalid dob
            {"dob": ["foo", "10", "2003"]},
            ("dob", "Invalid values.")
        ),
        (
            # empty picture
            {"picture": ""},
            ("picture", "This field is required.")
        ),
        (
            # empty phone
            {"phone": ""},
            ("phone", "This field is required.")
        ),
        (
            # non-numeric phone
            {"phone": "A12345"},
            ("phone", "Value is not a number.")
        ),
        (
            # non-numeric phone
            {"phone": "test"},
            ("phone", "Value is not a number.")
        ),
        (
            # empty email
            {"email": ""},
            ("email", "This field is required.")
        ),
        (
            # invalid email
            {"email": "test"},
            ("email", "Value is not an email.")
        ),
        (
            # Comments more than 10 chars
            {"comments": "This is a long comment."},
            ("comments", "Characters exceed limit of 10.")
        ),
))
def test_form_post_invalid(errors, expected, client, db, picture, html, find):
    data = {
        "name": "Test",
        "sex": "male",
        "gender": "yes",
        "ethnicity": "mixed",
        "dob": ["10", "10", "2000"],
        "phone": "12345",
        "email": "test@test",
        "comments": "Test",
    }
    data.update(errors)
    response = client.post(
        "/forms/profile",
        data=data,
        files={"picture": picture},
    )
    assert response.status_code == 200
    field_name, error_message = expected
    expected_html = html(
        f'<p class="govuk-error-message" id="{field_name}-error">'
            '<span class="govuk-visually-hidden">Error: </span>'
            f"{error_message}"
        "</p>"
    )
    error_p = find(response.text, "p", {"class": "govuk-error-message"})
    assert html(error_p) == expected_html


@patch("fast_gov_uk.forms.logger")
def test_log_form_post_valid(mock_logger, client):
    data = {"satisfaction": "satisfied"}
    response = client.post(
        "/forms/log_feedback",
        data=data,
    )
    assert response.status_code == 303
    logger_call_args = mock_logger.info.call_args
    assert logger_call_args == call(
        "Form: 'feedback' processed with: {'satisfaction': 'Satisfied'}."
    )


def test_email_form_post_valid(fast, db, client):
    data = {"satisfaction": "satisfied"}
    response = client.post(
        "/forms/email_feedback",
        data=data,
    )
    assert response.status_code == 303
    notify_call_args = fast.notify_client.send_email_notification.call_args
    assert notify_call_args == call(
        email_address='test@test.com',
        template_id='test',
        personalisation={
            'form_name': 'feedback',
            'form_data': '* satisfaction: Satisfied',
            'service_name': 'Fast-gov-uk test'
        }
    )


def test_api_form_post_valid(fast, db, client):
    data = {"satisfaction": "satisfied"}
    with patch("fast_gov_uk.forms._client") as mock_client:
        mock_post = Mock()
        mock_client.return_value = Mock(post=mock_post)
        response = client.post(
            "/forms/api_feedback",
            data=data,
        )
    assert response.status_code == 303
    assert mock_client.call_args == call("test_user", "test_password")
    assert mock_post.call_args == call(
        'https://test.com',
        data={
            'satisfaction': 'Satisfied',
            'form_name': 'feedback',
            'submitted_on': ANY,
        }
    )


def test_session_form_post_valid(fast, db, client):
    data = {"satisfaction": "satisfied"}
    response = client.post(
        "/forms/session_feedback",
        data=data,
        follow_redirects=True,
    )
    assert response.status_code == 200
    response = client.get("/session")
    feedback = response.json()["feedback"]
    assert feedback == {"satisfaction": "Satisfied"}


def test_questions_get(client):
    response = client.get("/wizards/mini_equality")
    assert response.status_code == 307
    response = client.get("/wizards/mini_equality/0")
    assert response.status_code == 200
    response = client.get("/wizards/mini_equality/1")
    assert response.status_code == 200
    response = client.get("/wizards/mini_equality/2")
    assert response.status_code == 200
    response = client.get("/wizards/mini_equality/3")
    assert response.status_code == 200


def test_question_no_permission(db, client):
    response = client.post(
        "/wizards/mini_equality/",
        data={"permission": "no"},
    )
    assert response.status_code == 303
    assert response.headers["Location"] == "/"


def test_question_permission(db, client):
    response = client.post(
        "/wizards/mini_equality/",
        data={"permission": "yes"},
    )
    assert response.status_code == 303
    assert response.headers["Location"] == "/wizards/mini_equality/1"


def test_questions_predicate_true(db, client):
    response = client.post(
        "/wizards/mini_equality/",
        data={"permission": "yes"},
    )
    assert response.status_code == 303
    assert response.headers["Location"] == "/wizards/mini_equality/1"


def test_questions_predicate_false(db, client):
    response = client.post(
        "/wizards/mini_equality",
        data={"permission": "no"},
        follow_redirects=True,
    )
    assert response.status_code == 200
    assert response.url == "http://testserver/"


def test_questions_valid(db, client):
    # First step
    response = client.post(
        "/wizards/mini_equality/",
        data={"permission": "yes"},
        follow_redirects=True,
    )
    assert response.status_code == 200
    # Second step
    response = client.post(
        response.url,
        data={"health": "yes"},
        follow_redirects=True,
    )
    assert response.status_code == 200
    # Third step
    response = client.post(
        response.url,
        data={"ability": "alot"},
        follow_redirects=True,
    )
    assert response.status_code == 200
    # Invalid input
    response = client.post(
        response.url,
        data={},
        follow_redirects=True,
    )
    assert response.status_code == 200
    # Still third step
    response = client.post(
        "/wizards/mini_equality/3",
        data={"sex": "skip", "gender": "skip"},
        follow_redirects=True,
    )
    assert response.status_code == 200
    assert response.url.path == "/"
    forms = db.t.forms()
    form = forms[0]
    assert form.name == "equality"
    form_json = asdict(form)
    form_data = form_json["data"]
    form_dict = json.loads(form_data)
    assert form_dict == {
        "permission": "Yes, answer the equality questions",
        "health": "Yes",
        "ability": "Yes, a lot",
        "sex": "Prefer not to say",
        "gender": "Prefer not to say"
    }
    # Invalid step
    response = client.get(
        "/wizards/mini_equality/20",
        follow_redirects=True,
    )
    assert response.status_code == 404



def test_error_summary(html, find):
    form = session_feedback(data={})
    expected = html(
        # The error summary - containine a link to the Radio field
        '<div class="govuk-error-summary" data-module="govuk-error-summary">'
            '<div role="alert">'
                '<h2 class="govuk-error-summary__title">There is a problem</h2>'
                '<div class="govuk-error-summary__body">'
                    '<ul class="govuk-list govuk-error-summary__list">'
                        '<li><a class="govuk-link" href="#satisfaction">How satisfied did you feel about the service?</a></li>'
                    "</ul>"
                "</div>"
            "</div>"
        "</div>"
    )
    form_html = html(form)
    error_summary = find(form_html, "div", {"class": "govuk-error-summary"})
    assert html(error_summary) == expected


def test_no_fields_at_root():
    with pytest.raises(ValueError):
        forms.Form("test", "test")


@patch("fast_gov_uk.forms.httpx")
def test_client(mock_httpx):
    _ = forms._client("test", "test")
    auth = mock_httpx.BasicAuth
    assert auth.call_args == call(username="test", password="test")
    client = mock_httpx.Client
    assert client.called
